import contextlib
import sqlite3
import time
from pathlib import Path

from eth.db.backends.base import BaseDB

# from vyper.utils import timeit

# poor man's constant
_ONE_MONTH = 30 * 24 * 3600


def get_current_time() -> float:
    return time.time()


class SqliteCache(BaseDB):
    """
    Cache which uses sqlite as its backing store which conforms to the
    interface of BaseDB.
    """

    _GLOBAL = None

    _debug = False

    _PRAGMA_CMDS = [
        """
        -- tuning
        pragma journal_mode = wal;
        """,
        """
        pragma temp_store = memory;
        """,
        """
        -- it's ok to lose some of the wal, this is just a cache
        pragma synchronous = normal;
        """,
        """
        -- https://sqlite.org/forum/forumpost/3ce1ee76242cfb29
        /* "I'm of the opinion that you should never use mmap"
           - Richard Hipp
        */
        -- pragma mmap_size = 30000000000;
        -- not sure if these help or hurt things:
        -- pragma auto_vacuum = incremental;
        -- pragma page_size = 512;
        pragma cache_size = 10000;
        """,
    ]

    _CREATE_CMDS = [
        """
        -- initialize schema
        CREATE TABLE IF NOT EXISTS kv_store (
            key TEXT PRIMARY KEY, value BLOB, expires_at float
        );
        """,
        """
        CREATE INDEX IF NOT EXISTS expires_at_index ON kv_store(expires_at);
        """,
    ]

    # flush at least once every second
    _FLUSH_INTERVAL = 1.0

    def __init__(self, db_path: Path | str, ttl: float = _ONE_MONTH) -> None:
        if db_path != ":memory:":  # sqlite magic path
            db_path = Path(db_path)
            db_path.parent.mkdir(parents=True, exist_ok=True)

        # short timeout so that we don't get stuck waiting for
        # OperationalError on exit (e.g. in __del__()), but long enough
        # so that we don't spinlock in acquire lock. 10ms seems like
        # a good balance.
        # note that informal benchmarks performed on my machine(TM) had
        # the following results:
        # __getitem__: 45us
        # __getitem__ (from wal): 120us
        # __setitem__: 300us
        # gc: 40us
        # flush: 6000us
        TIMEOUT = 0.010  # 10ms

        self.db: sqlite3.Connection = sqlite3.connect(
            db_path, timeout=TIMEOUT, isolation_level=None
        )

        # cache the cursor
        self._cursor = self.db.cursor()

        # initialize other fields
        self._last_flush = get_current_time()
        self._expiry_updates: list[tuple[float, bytes]] = []

        # ttl of cache entries in seconds
        self.ttl: float = float(ttl)

        # executescript does not work in multiprocess environment
        for cmd in self._PRAGMA_CMDS:
            self._cursor.execute(cmd)
        with self.acquire_write_lock():
            for cmd in self._CREATE_CMDS:
                self._cursor.execute(cmd)

        self.gc()

    def gc(self):
        with self.acquire_write_lock():
            current_time = get_current_time()
            self._cursor.execute(
                "DELETE FROM kv_store WHERE expires_at < ?", (current_time,)
            )
            # batch writes in here
            self._flush(nolock=True)

    def __del__(self):
        # try to flush but if we fail, no worries. these are optional
        # operations.
        try:
            self._flush()
        except Exception:
            pass

    def _flush_condition(self):
        if len(self._expiry_updates) >= 1000:
            return True
        next_flush = self._last_flush + self._FLUSH_INTERVAL
        if get_current_time() > next_flush:
            return True
        return False

    # @timeit("FLUSH")
    def _flush(self, nolock=False):
        # set nolock=True if the caller has already acquired a lock.
        if len(self._expiry_updates) == 0:
            self._last_flush = get_current_time()
            return

        acquire_lock = contextlib.nullcontext if nolock else self.acquire_write_lock
        with acquire_lock():
            query_string = """
            UPDATE kv_store
                SET expires_at=?
                WHERE key=?
            """
            self._cursor.executemany(query_string, self._expiry_updates)

        self._expiry_updates = []
        self._last_flush = get_current_time()

    @contextlib.contextmanager
    def acquire_write_lock(self):
        count = 0
        while True:
            try:
                self._cursor.execute("BEGIN IMMEDIATE")
                break
            except sqlite3.OperationalError as e:
                count += 1
                # deadlock scenario, should not happen
                if count > 1000:
                    msg = "deadlock encountered! this means there is a bug in"
                    msg += " titanoboa. please report this on the titanoboa"
                    msg += " issue tracker on github. in the meantime as a"
                    msg += " workaround, try disabling the sqlite cache."
                    raise Exception(msg) from e
                # sleep 100 micros, roughly the time for a write query to
                # complete.
                # keep in mind that time.sleep takes 50us+:
                # https://github.com/python/cpython/issues/125997.
                time.sleep(1e-4)
                continue
        try:
            yield
            self._cursor.execute("COMMIT")
        except Exception as e:
            if self._debug:
                # if we are in tests, raise the exception
                raise Exception("fail") from e
            # this shouldn't really happen, but it could happen if there
            # is some concurrent, non-boa write to the database, so just
            # roll back (fail "gracefully") and move on.
            self._cursor.execute("ROLLBACK")

    @classmethod
    # Creates db as a singleton class variable
    def create(cls, *args, **kwargs):
        if cls._GLOBAL is None:
            cls._GLOBAL = cls(*args, **kwargs)
        return cls._GLOBAL

    def get_expiry_ts(self) -> float:
        current_time = get_current_time()
        return current_time + self.ttl

    # @timeit("CACHE HIT")
    def __getitem__(self, key: bytes) -> bytes:
        query_string = """
        SELECT value, expires_at FROM kv_store
            WHERE key=?
        """
        # with timeit("CACHE HIT"):
        res = self._cursor.execute(query_string, (key,)).fetchone()
        if res is None:
            raise KeyError(key)

        val, expires_at = res

        # to reduce contention, instead of updating the expiry every
        # time, batch the expiry updates.
        if expires_at - get_current_time() > self.ttl / 100:
            new_expiry_ts = self.get_expiry_ts()
            self._expiry_updates.append((new_expiry_ts, key))
            if self._flush_condition():
                self._flush()

        return val

    # @timeit("CACHE MISS")
    def __setitem__(self, key: bytes, value: bytes) -> None:
        # with timeit("CACHE MISS"):
        with self.acquire_write_lock():
            query_string = """
            INSERT INTO kv_store(key, value, expires_at) VALUES (?,?,?)
              ON CONFLICT DO UPDATE
              SET value=excluded.value,
                  expires_at=excluded.expires_at
            """
            expiry_ts = self.get_expiry_ts()
            self._cursor.execute(query_string, (key, value, expiry_ts))

            # we already have a lock, batch in some writes if needed
            if self._flush_condition():
                self._flush(nolock=True)

    def _exists(self, key: bytes) -> bool:
        query_string = "SELECT count(*) FROM kv_store WHERE key=?"
        (res,) = self._cursor.execute(query_string, (key,)).fetchone()
        return bool(res)

    def __delitem__(self, key: bytes) -> None:
        with self.acquire_write_lock():
            res = self._cursor.execute("DELETE FROM kv_store WHERE key=?", (key,))
        if res.rowcount == 0:
            raise KeyError(key)



import re
import time
from dataclasses import dataclass
from datetime import timedelta
from typing import Optional

from boa.rpc import json
from boa.util.abi import Address
from boa.verifiers import ContractVerifier, VerificationResult

try:
    from requests_cache import CachedSession

    def filter_fn(response):
        return response.ok and _is_success_response(response.json())

    SESSION = CachedSession(
        "~/.cache/titanoboa/explorer_cache",
        filter_fn=filter_fn,
        allowable_codes=[200],
        cache_control=True,
        expire_after=3600 * 6,
        stale_if_error=True,
        stale_while_revalidate=True,
    )
except ImportError:
    from requests import Session

    SESSION = Session()

DEFAULT_ETHERSCAN_URI = "https://api.etherscan.io/v2/api"  # Etherscan API v2
DEFAULT_CHAIN_ID = 1  # Mainnet
VERSION_RE = re.compile(r"v(\d+\.\d+\.\d+)(\+commit.*)?")


@dataclass
class Etherscan(ContractVerifier[str]):
    uri: Optional[str] = DEFAULT_ETHERSCAN_URI
    api_key: Optional[str] = None
    num_retries: int = 10
    backoff_ms: int | float = 400.0
    backoff_factor: float = 1.1  # 1.1**10 ~= 2.59
    timeout = timedelta(minutes=2)
    chain_id: Optional[int] = DEFAULT_CHAIN_ID

    # Setter
    def set_chain_id(self, value: Optional[int]):
        if not isinstance(value, int) or value <= 0:
            raise ValueError("Chain ID must be a positive integer.")
        self.chain_id = value

    # Methods
    def verify(
        self,
        address: Address,
        contract_name: str,
        solc_json: dict,
        constructor_calldata: bytes,
        license_type: str = "1",
        wait: bool = False,
    ) -> Optional["VerificationResult[str]"]:
        """
        Verify the Vyper contract on Etherscan.
        :param address: The address of the contract.
        :param contract_name: The name of the contract.
        :param solc_json: The solc_json output of the Vyper compiler.
        :param constructor_calldata: The calldata for the contract constructor.
        :param chain_id: The ID of the chain where the contract is deployed.
        :param license_type: The license to use for the contract. Defaults to "none".
        :param wait: Whether to return a VerificationResult immediately
                     or wait for verification to complete. Defaults to False
        """
        api_key = self.api_key or ""
        output_selection = solc_json["settings"]["outputSelection"]
        contract_file = next(k for k, v in output_selection.items() if "*" in v)
        compiler_version = solc_json["compiler_version"]
        version_match = re.match(VERSION_RE, compiler_version)
        if not version_match:
            raise ValueError(f"Failed to extract Vyper version from {compiler_version}")

        data = {
            "module": "contract",
            "action": "verifysourcecode",
            "apikey": api_key,
            "chainid": self.chain_id,
            "codeformat": "vyper-json",
            "sourceCode": json.dumps(solc_json),
            "constructorArguments": constructor_calldata.hex(),
            "contractaddress": address,
            "contractname": f"{contract_file}:{contract_name}",
            "compilerversion": f"vyper:{version_match.group(1)}",
            "optimizationUsed": "1",
            "licenseType": license_type,
        }

        def verification_created():
            # we need to retry until the contract is found by Etherscan
            response = SESSION.post(self.uri, data=data)
            response.raise_for_status()
            response_json = response.json()
            if response_json.get("status") == "1":
                return response_json["result"]
            if (
                response_json.get("message") == "NOTOK"
                and "Unable to locate ContractCode" not in response_json["result"]
            ):
                raise ValueError(f"Failed to verify: {response_json['result']}")
            print(
                f"Verification could not be created yet: {response_json['result']}. Retrying..."
            )
            return None

        etherscan_guid = self._wait_until(
            verification_created, timedelta(minutes=2), timedelta(seconds=5), 1.1
        )
        print(f"Verification started with etherscan_guid {etherscan_guid}")
        if not wait:
            return VerificationResult(etherscan_guid, self)

        self.wait_for_verification(etherscan_guid)
        return None

    def wait_for_verification(self, etherscan_guid: str) -> None:
        """
        Waits for the contract to be verified on Etherscan.
        :param etherscan_guid: The unique ID of the contract verification.
        """
        self._wait_until(
            lambda: self.is_verified(etherscan_guid),
            self.timeout,
            self.backoff,
            self.backoff_factor,
        )
        print("Contract verified!")

    @property
    def backoff(self):
        return timedelta(milliseconds=self.backoff_ms)

    def is_verified(self, etherscan_guid: str) -> bool:
        api_key = self.api_key or ""
        url = f"{self.uri}?module=contract&action=checkverifystatus"
        url += f"&guid={etherscan_guid}&apikey={api_key}&chainid={self.chain_id}"

        response = SESSION.get(url)
        response.raise_for_status()
        response_json = response.json()
        if response_json.get("message") == "NOTOK":
            if "Already Verified" in response_json["result"]:
                print("Contract already verified")
                return True
            if "Pending in queue" not in response_json["result"]:
                raise ValueError(f"Failed to verify: {response_json['result']}")
        return response_json.get("status") == "1"

    def __post_init__(self):
        if self.uri is None:
            self.uri = DEFAULT_ETHERSCAN_URI

        # call set_chain_id for validation
        self.set_chain_id(self.chain_id)

    def _fetch(self, **params) -> dict:
        """
        Fetch data from Etherscan API.
        Offers a simple caching mechanism to avoid redundant queries.
        Retries if rate limit is reached.
        :param num_retries: Number of retries
        :param backoff_ms: Backoff in milliseconds
        :param params: Additional query parameters
        :return: JSON response
        """
        params = {**params, "apiKey": self.api_key, "chainid": self.chain_id}
        for i in range(self.num_retries):
            res = SESSION.get(self.uri, params=params)
            res.raise_for_status()
            data = res.json()
            if not _is_rate_limited(data):
                break

            f = self.backoff_factor**i
            seconds = self.backoff_ms / 1000
            time.sleep(f * seconds)

        if not _is_success_response(data):
            raise ValueError(f"Failed to retrieve data from API: {data}")

        return data

    def fetch_abi(self, address: str):
        # resolve implementation address if `address` is a proxy contract
        address = self._resolve_implementation_address(address)

        # fetch ABI of `address`
        params = dict(module="contract", action="getabi", address=address)
        data = self._fetch(**params)

        return json.loads(data["result"].strip())

    # fetch the address of a contract; resolves at most one layer of
    # indirection if the address is a proxy contract.
    def _resolve_implementation_address(self, address: str) -> str:
        params = dict(module="contract", action="getsourcecode", address=address)
        data = self._fetch(**params)
        source_data = data["result"][0]

        # check if the contract is a proxy
        if int(source_data["Proxy"]) == 1:
            return source_data["Implementation"]
        else:
            return address


_etherscan = Etherscan()


def get_etherscan():
    return _etherscan


def _set_etherscan(etherscan: Etherscan):
    global _etherscan
    _etherscan = etherscan


def _is_success_response(data: dict) -> bool:
    return data.get("status") == "1"


def _is_rate_limited(data: dict) -> bool:
    """
    Check if the response is rate limited. Possible error messages:
    - Max calls per sec rate limit reached (X/sec)
    - Max rate limit reached, please use API Key for higher rate limit
    - Max rate limit reached
    :param data: Etherscan API response
    :return: True if rate limited, False otherwise
    """
    return "rate limit" in data.get("result", "") and data.get("status") == "0"



import contextlib
import sys
import textwrap
import warnings
from importlib.abc import MetaPathFinder
from importlib.machinery import SourceFileLoader
from importlib.util import spec_from_loader
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union

import vvm
import vyper
import vyper.ir.compile_ir as compile_ir
from packaging.version import Version
from vvm.utils.versioning import _pick_vyper_version, detect_version_specifier_set
from vyper.ast.parse import parse_to_ast
from vyper.cli.vyper_compile import get_search_paths
from vyper.compiler.input_bundle import CompilerInput, FileInput, FilesystemInputBundle
from vyper.compiler.phases import CompilerData
from vyper.compiler.settings import Settings, anchor_settings
from vyper.semantics.analysis.imports import resolve_imports
from vyper.semantics.analysis.module import analyze_module
from vyper.semantics.types.module import ModuleT
from vyper.utils import sha256sum

from boa.contracts.abi.abi_contract import ABIContractFactory
from boa.contracts.vvm.vvm_contract import VVMDeployer
from boa.contracts.vyper.vyper_contract import (
    VyperBlueprint,
    VyperContract,
    VyperDeployer,
)
from boa.environment import Env
from boa.explorer import Etherscan, get_etherscan
from boa.rpc import json
from boa.util.abi import Address
from boa.util.disk_cache import DiskCache

if TYPE_CHECKING:
    from vyper.semantics.analysis.base import ImportInfo

_Contract = Union[VyperContract, VyperBlueprint]


_disk_cache = None
_search_path = None


def set_search_paths(path: list[str]):
    global _search_path
    _search_path = path


def set_search_path(paths: list[str]):
    warnings.warn(
        DeprecationWarning("set_search_path is deprecated, use set_search_paths.")
    )
    set_search_paths(paths)


def set_cache_dir(cache_dir="~/.cache/titanoboa"):
    global _disk_cache
    if cache_dir is None:
        _disk_cache = None
        return
    compiler_version = f"{vyper.__version__}.{vyper.__commit__}"
    _disk_cache = DiskCache(cache_dir, compiler_version)


def disable_cache():
    set_cache_dir(None)


set_cache_dir()  # enable caching, by default!


class BoaImporter(MetaPathFinder):
    def find_spec(self, fullname, path, target=None):
        path = Path(fullname.replace(".", "/")).with_suffix(".vy")

        for prefix in sys.path:
            to_try = Path(prefix) / path

            if to_try.exists():
                loader = BoaLoader(fullname, str(to_try))
                return spec_from_loader(fullname, loader)


class BoaLoader(SourceFileLoader):
    def get_code(self, fullname):
        # importlib docs say to return None, but that triggers an `ImportError`
        return ""

    def create_module(self, spec):
        ret = load_partial(self.path)

        # comply with PEP-302:
        ret.__name__ = spec.name
        ret.__file__ = self.path
        ret.__loader__ = self
        ret.__package__ = spec.name.rpartition(".")[0]
        return ret


sys.meta_path.append(BoaImporter())


def hash_input(compiler_input: CompilerInput) -> str:
    return compiler_input.sha256sum


# compute a fingerprint for a module which changes if any of its
# dependencies change
# TODO consider putting this in its own module
# TODO: refactor with new machinery in vyper 0.4.2+
def get_module_fingerprint(
    module_t: ModuleT, seen: dict["ImportInfo", str] = None
) -> str:
    seen = seen or {}
    fingerprints = []
    for stmt in module_t.import_stmts:
        import_info = stmt._metadata["import_info"]
        if id(import_info) not in seen:
            if isinstance(import_info.typ, ModuleT):
                fingerprint = get_module_fingerprint(import_info.typ, seen)
            else:
                fingerprint = hash_input(import_info.compiler_input)
            seen[id(import_info)] = fingerprint
        fingerprint = seen[id(import_info)]
        fingerprints.append(fingerprint)
    fingerprints.append(module_t._module.source_sha256sum)

    return sha256sum("".join(fingerprints))


def compiler_data(
    source_code: str,
    contract_name: str | None,
    filename: str | Path,
    deployer=None,
    **kwargs,
) -> CompilerData:
    global _disk_cache, _search_path

    path = Path(filename)
    resolved_path = Path(filename).resolve(strict=False)

    file_input = FileInput(
        contents=source_code, source_id=-1, path=path, resolved_path=resolved_path
    )

    search_paths = get_search_paths(_search_path)
    input_bundle = FilesystemInputBundle(search_paths)

    settings = Settings(**kwargs)
    ret = CompilerData(file_input, input_bundle, settings)
    if _disk_cache is None:
        return ret

    with anchor_settings(ret.settings):
        # note that this actually parses and analyzes all dependencies,
        # even if they haven't changed. an optimization would be to
        # somehow convince vyper (in ModuleAnalyzer) to get the module_t
        # from the cache.
        module_t = ret.annotated_vyper_module._metadata["type"]
    fingerprint = get_module_fingerprint(module_t)

    def get_compiler_data():
        with anchor_settings(ret.settings):
            # force compilation to happen so DiskCache will cache the compiled artifact:
            _ = ret.bytecode, ret.bytecode_runtime

            # workaround since CompilerData does not compute source_map
            if not hasattr(ret, "source_map"):
                # cache source map
                ret.source_map = _compute_source_map(ret)

        return ret

    assert isinstance(deployer, type) or deployer is None
    deployer_id = repr(deployer)  # a unique str identifying the deployer class
    cache_key = str((contract_name, filename, fingerprint, kwargs, deployer_id))

    ret = _disk_cache.caching_lookup(cache_key, get_compiler_data)

    if not hasattr(ret, "source_map"):
        # invalidate so it will be cached on the next run
        _disk_cache.invalidate(cache_key)

        # compute source map so it's available downstream
        ret.source_map = _compute_source_map(ret)

    return ret


def _compute_source_map(compiler_data: CompilerData) -> Any:
    _, source_map = compile_ir.assembly_to_evm(compiler_data.assembly_runtime)
    return source_map


def load(filename: str | Path, *args, **kwargs) -> _Contract:  # type: ignore
    name = Path(filename).stem
    # TODO: investigate if we can just put name in the signature
    if "name" in kwargs:
        name = kwargs.pop("name")
    with open(filename) as f:
        return loads(f.read(), *args, name=name, **kwargs, filename=filename)


def loads(
    source_code,
    *args,
    as_blueprint=False,
    name=None,
    filename=None,
    compiler_args=None,
    no_vvm=False,
    **kwargs,
):
    d = loads_partial(
        source_code, name, filename=filename, compiler_args=compiler_args, no_vvm=no_vvm
    )
    if as_blueprint:
        return d.deploy_as_blueprint(contract_name=name, **kwargs)
    else:
        return d.deploy(*args, contract_name=name, **kwargs)


def load_abi(filename: str, *args, name: str = None, **kwargs) -> ABIContractFactory:
    if name is None:
        name = Path(filename).stem
    # TODO: pass filename to ABIContractFactory
    with open(filename) as fp:
        return loads_abi(fp.read(), *args, name=name, **kwargs)


def loads_abi(json_str: str, *args, name: str = None, **kwargs) -> ABIContractFactory:
    return ABIContractFactory.from_abi_dict(json.loads(json_str), name, *args, **kwargs)


# load from .vyi file.
# NOTE: substantially same interface as load_abi and loads_abi, consider
# fusing them into load_interface?
def load_vyi(filename: str, name: str = None) -> ABIContractFactory:
    if name is None:
        name = Path(filename).stem
    with open(filename) as fp:
        return loads_vyi(fp.read(), name=name, filename=filename)


# load interface from .vyi file string contents.
# NOTE: since vyi files can be compiled in 0.4.1, this codepath can probably
# be refactored to use CompilerData (or straight loads_partial)
def loads_vyi(source_code: str, name: str = None, filename: str = None):
    global _search_path

    ast = parse_to_ast(source_code, is_interface=True)

    if name is None:
        name = "VyperContract.vyi"

    search_paths = get_search_paths(_search_path)
    input_bundle = FilesystemInputBundle(search_paths)

    # cf. CompilerData._resolve_imports
    if filename is not None:
        ctx = input_bundle.search_path(Path(filename).parent)
    else:
        ctx = contextlib.nullcontext()
    with ctx:
        _ = resolve_imports(ast, input_bundle)

    module_t = analyze_module(ast)
    abi = module_t.interface.to_toplevel_abi_dict()
    return ABIContractFactory(name, abi, filename=filename)


def loads_partial(
    source_code: str,
    name: str = None,
    filename: str | Path | None = None,
    dedent: bool = True,
    compiler_args: dict = None,
    no_vvm: bool = False,
) -> VyperDeployer:
    if filename is None:
        filename = "<unknown>"

    if dedent:
        source_code = textwrap.dedent(source_code)

    if not no_vvm:
        specifier_set = detect_version_specifier_set(source_code)
        # Use VVM only if the installed version is not in the specifier set
        if specifier_set is not None and not specifier_set.contains(vyper.__version__):
            version = _pick_vyper_version(specifier_set)
            filename = str(filename)  # help mypy
            return _loads_partial_vvm(source_code, version, name, filename)

    compiler_args = compiler_args or {}

    deployer_class = _get_default_deployer_class()
    data = compiler_data(source_code, name, filename, deployer_class, **compiler_args)
    return deployer_class(data, filename=filename)


def load_partial(filename: str, compiler_args=None):
    with open(filename) as f:
        return loads_partial(
            f.read(), name=filename, filename=filename, compiler_args=compiler_args
        )


def _loads_partial_vvm(
    source_code: str,
    version: Version,
    name: Optional[str],
    filename: str,
    base_path=None,
):
    global _disk_cache

    if base_path is None:
        base_path = Path(".")

    # install the requested version if not already installed
    vvm.install_vyper(version=version)

    def _compile():
        return vvm.compile_source(
            source_code, vyper_version=version, base_path=base_path
        )

    # separate _handle_output and _compile so that we don't trample
    # name and filename in the VVMDeployer from separate invocations
    # (with different values for name+filename).
    def _handle_output(compiled_src):
        compiler_output = compiled_src["<stdin>"]
        return VVMDeployer.from_compiler_output(
            compiler_output, name=name, filename=filename
        )

    # Ensure the cache is initialized
    if _disk_cache is None:
        return _handle_output(_compile())

    # Generate a unique cache key
    cache_key = f"{source_code}:{version}"

    # Check the cache and return the result if available
    ret = _disk_cache.caching_lookup(cache_key, _compile)

    # backwards compatibility: old versions of boa returned a VVMDeployer.
    # here we detect the case and invalidate the cache so it can recompile.
    if isinstance(ret, VVMDeployer):
        _disk_cache.invalidate(cache_key)
        ret = _disk_cache.caching_lookup(cache_key, _compile)

    return _handle_output(ret)


def from_etherscan(
    address: Any,
    name: str = None,
    uri: str = None,
    api_key: str = None,
    chain_id: int = None,
):
    addr = Address(address)

    if uri is not None or api_key is not None:
        etherscan = Etherscan(uri, api_key)
    else:
        etherscan = get_etherscan()

    if chain_id is None:
        # default behavior: use the chain id of the global env
        chain_id = Env.get_singleton().evm.patch.chain_id

    # Set the chain ID for the Etherscan instance
    etherscan.set_chain_id(chain_id)

    abi = etherscan.fetch_abi(addr)
    return ABIContractFactory.from_abi_dict(abi, name=name).at(addr)


def _get_default_deployer_class():
    env = Env.get_singleton()
    if hasattr(env, "deployer_class"):
        return env.deployer_class
    return VyperDeployer


__all__ = []  # type: ignore



from functools import cached_property
from typing import Optional

from boa.contracts.abi.abi_contract import ABIContract, ABIContractFactory, ABIFunction
from boa.environment import Env
from boa.util.abi import Address
from boa.util.eip5202 import generate_blueprint_bytecode


class VVMBlueprint(ABIContract):
    def __init__(self, deployer: "VVMDeployer", address: Address):
        name = deployer.name or "<unknown>"  # help mypy
        super().__init__(
            name,
            abi=[],
            address=address,
            filename=deployer.filename,
            functions=[],
            events=[],
        )
        self._deployer = deployer

    @property
    def deployer(self):
        return self._deployer


class VVMDeployer:
    """
    A deployer that uses the Vyper Version Manager (VVM).
    This allows deployment of contracts written in older versions of Vyper that
    can interact with new versions using the ABI definition.
    """

    def __init__(self, abi, bytecode, name, filename):
        """
        Initialize a VVMDeployer instance.
        :param abi: The contract's ABI.
        :param bytecode: The contract's bytecode.
        :param filename: The filename of the contract.
        """
        self.abi: dict = abi
        self.bytecode: bytes = bytecode
        self.name: Optional[str] = name
        self.filename: str = filename

    @classmethod
    def from_compiler_output(cls, compiler_output, name, filename):
        abi = compiler_output["abi"]
        bytecode_nibbles = compiler_output["bytecode"]
        bytecode = bytes.fromhex(bytecode_nibbles.removeprefix("0x"))
        return cls(abi, bytecode, name, filename)

    @cached_property
    def factory(self):
        return ABIContractFactory.from_abi_dict(
            self.abi, name=self.name, filename=self.filename
        )

    @cached_property
    def constructor(self):
        for t in self.abi:
            if t["type"] == "constructor":
                return ABIFunction(t, contract_name=self.filename)
        return None

    def deploy(self, *args, contract_name=None, env=None, **kwargs):
        encoded_args = b""
        if self.constructor is not None:
            encoded_args = self.constructor.prepare_calldata(*args)
        elif len(args) > 0:
            raise ValueError(f"No constructor, but args were provided: {args}")

        if env is None:
            env = Env.get_singleton()

        address, computation = env.deploy(
            bytecode=self.bytecode + encoded_args, **kwargs
        )

        # set nowarn=True. if there was a problem in the deploy, it will
        # be caught at computation.is_error, so the warning is redundant
        # (and annoying!)
        ret = self.at(address, nowarn=True)
        if contract_name is not None:
            # override contract name
            ret.contract_name = contract_name
        ret._computation = computation

        if computation.is_error:
            ret.handle_error(computation)

        return ret

    def deploy_as_blueprint(self, env=None, blueprint_preamble=None, **kwargs):
        """
        Deploy a new blueprint from this contract.
        :param blueprint_preamble: The preamble to use for the blueprint.
        :param env: The environment to deploy the blueprint in.
        :param kwargs: Keyword arguments to pass to the environment `deploy_code` method.
        :returns: A contract instance.
        """
        if env is None:
            env = Env.get_singleton()

        blueprint_bytecode = generate_blueprint_bytecode(
            self.bytecode, blueprint_preamble
        )
        address, computation = env.deploy(bytecode=blueprint_bytecode, **kwargs)

        ret = VVMBlueprint(self, address)

        if computation.is_error:
            ret.handle_error(computation)

        env.register_blueprint(self.bytecode, ret)
        return ret

    def __call__(self, *args, **kwargs):
        return self.deploy(*args, **kwargs)

    def at(self, address, nowarn=False):
        return self.factory.at(address, nowarn=nowarn)



import io
import re
import tokenize
from functools import lru_cache
from typing import Any, Optional

import vyper.ast as vy_ast


@lru_cache(maxsize=128)
def _get_tokens_for_file(source_code: str):
    """Tokenize entire file and cache the result with line number index."""
    tokens = list(tokenize.generate_tokens(io.StringIO(source_code).readline))

    # Create an index from line number to tokens on that line
    line_to_tokens = {}
    for token in tokens:
        line_num = token.start[0]
        if line_num not in line_to_tokens:
            line_to_tokens[line_num] = []
        line_to_tokens[line_num].append(token)

    return line_to_tokens


# loosely, match # `@dev asdf...` or `dev: asdf...`
REASON_PATTERN = re.compile(r"#\s*@?(\w+):?\s+(.*)")


def _extract_reason(comment: str) -> Any:
    m = REASON_PATTERN.match(comment)
    if m is not None:
        return m.group(1, 2)
    return None


# extract the dev revert reason at a given line.
# somewhat heuristic.
def reason_at(
    source_code: str, lineno: int, end_lineno: int
) -> Optional[tuple[str, str]]:
    line_to_tokens = _get_tokens_for_file(source_code)

    # Only check tokens in the specified line range
    for line_num in range(lineno, end_lineno + 1):
        for token in line_to_tokens.get(line_num, []):
            if token.type == tokenize.COMMENT:
                reason = _extract_reason(token.string)
                if reason:
                    return reason
    return None


# TODO: maybe move this into boa/profiling.py
def get_fn_name_from_lineno(ast_map: dict, path: str, lineno: int) -> str:
    for node in ast_map.values():
        if node.lineno == lineno and node.module_node.resolved_path == path:
            fn_node = get_fn_ancestor_from_node(node)
            if fn_node:
                return fn_node.name
    return ""


def get_fn_ancestor_from_node(node):
    if node is None:
        return None

    if isinstance(node, vy_ast.FunctionDef):
        return node

    return node.get_ancestor(vy_ast.FunctionDef)



from collections import namedtuple
from dataclasses import dataclass
from typing import Any, NamedTuple

from boa.util.abi import Address, abi_decode


@dataclass
class RawLogEntry:
    log_id: int  # internal py-evm log id, for ordering purposes
    address: str  # canonical address
    topics: list[int]  # list of topics
    data: bytes  # list of encoded args


def decode_log(
    addr: Address, event_abi_for: dict[int, dict], log_entry: RawLogEntry
) -> NamedTuple:
    # decode a log given (json) event abi. constructs a fresh namedtuple
    # type to put the data into, with a schema like
    # namedtuple(event_name, ["address", *event_fields])
    # TODO: resolve these cyclic imports. probably, `_parse_complex` and
    # `_abi_from_json` belong in an abi utility module of some sort
    from boa.contracts.abi.abi_contract import (
        _abi_from_json,
        _format_abi_type,
        _parse_complex,
    )

    assert addr.canonical_address == log_entry.address
    event_hash = log_entry.topics[0]

    # map from event id to event abi for the topic
    if event_hash not in event_abi_for:
        # our abi is wrong, we can't decode it. fail loudly.
        msg = f"can't find event with hash {hex(event_hash)} in abi"
        msg += f" (possible events: {event_abi_for})"
        raise ValueError(msg)

    event_abi = event_abi_for[event_hash]

    topic_abis = []
    arg_abis = []

    # add `address` to the tuple. this is prevented from being an
    # actual fieldname in vyper and solidity since it is a reserved keyword
    # in both languages. if for some reason some abi actually has a field
    # named `address`, it will be renamed by namedtuple(rename=True).
    tuple_names = ["address"]

    for item_abi in event_abi["inputs"]:
        is_topic = item_abi["indexed"]
        assert isinstance(is_topic, bool)
        if not is_topic:
            arg_abis.append(item_abi)
        else:
            topic_abis.append(item_abi)

        tuple_names.append(item_abi["name"])

    # to investigate: is this a hotspot?
    tuple_typ = namedtuple(event_abi["name"], tuple_names, rename=True)  # type: ignore[misc]

    decoded_topics = []
    for topic_abi, topic_int in zip(topic_abis, log_entry.topics[1:]):
        # convert to bytes for abi decoder
        encoded_topic = topic_int.to_bytes(32, "big")
        decoded_topics.append(abi_decode(_abi_from_json(topic_abi), encoded_topic))
    args_selector = _format_abi_type([_abi_from_json(arg_abi) for arg_abi in arg_abis])

    decoded_args = abi_decode(args_selector, log_entry.data)

    topics_ix = 0
    args_ix = 0

    xs: list[Any] = [Address(log_entry.address)]

    # re-align the evm topic + args lists with the way they appear in the
    # abi ex. Transfer(indexed address, address, indexed address)
    for item_abi in event_abi["inputs"]:
        is_topic = item_abi["indexed"]
        if is_topic:
            abi = topic_abis[topics_ix]
            topic = decoded_topics[topics_ix]
            # topic abi is currently never complex, but use _parse_complex
            # as future-proofing mechanism
            xs.append(_parse_complex(abi, topic))
            topics_ix += 1
        else:
            abi = arg_abis[args_ix]
            arg = decoded_args[args_ix]
            xs.append(_parse_complex(abi, arg))
            args_ix += 1

    return tuple_typ(*xs)



# wrapper module around whatever encoder we are using
from collections import deque
from typing import Annotated, Any

from eth.codecs.abi import nodes
from eth.codecs.abi.decoder import DecodeError, Decoder
from eth.codecs.abi.encoder import Encoder
from eth.codecs.abi.exceptions import ABIError
from eth.codecs.abi.nodes import ABITypeNode
from eth.codecs.abi.parser import Parser
from eth_typing import Address as PYEVM_Address
from eth_utils import to_canonical_address, to_checksum_address

from boa.util.lrudict import lrudict

_parsers: dict[str, ABITypeNode] = {}


# inherit from `str` so that users can compare with regular hex string
# addresses
class Address(str):
    # converting between checksum and canonical addresses is a hotspot;
    # this class contains both and caches recently seen conversions
    __slots__ = ("canonical_address",)
    _cache = lrudict(1024)

    canonical_address: Annotated[PYEVM_Address, "canonical address"]

    def __new__(cls, address):
        if isinstance(address, Address):
            return address

        try:
            return cls._cache[address]
        except KeyError:
            pass

        checksum_address = to_checksum_address(address)
        self = super().__new__(cls, checksum_address)
        self.canonical_address = to_canonical_address(address)
        cls._cache[address] = self
        return self

    def __repr__(self):
        checksum_addr = super().__repr__()
        return f"Address({checksum_addr})"


class _ABIEncoder(Encoder):
    """
    Custom encoder that extracts the address from an `Address` object
    and passes the result to the base encoder.
    """

    @classmethod
    def visit_AddressNode(cls, node: nodes.AddressNode, value) -> bytes:
        value = getattr(value, "address", value)
        return super().visit_AddressNode(node, value)


class _ABIDecoder(Decoder):
    """
    Custom decoder that wraps address results into an `Address` object.
    """

    @classmethod
    def visit_AddressNode(
        cls, node: nodes.AddressNode, value: bytes, checksum: bool = True, **kwargs: Any
    ) -> "Address":
        ret = super().visit_AddressNode(node, value)
        return Address(ret)

    @classmethod
    def visit_TupleNode(cls, node: nodes.TupleNode, value: bytes, **kwargs) -> tuple:
        """
        Monkey patch to workaround upstream bug in visit_TupleNode.
        See https://github.com/skellet0r/eth-stdlib/pull/22
        """
        size = sum((elem.width for elem in node.ctypes))
        # value size should be >= the sum of the length of its components
        if len(value) < size:
            raise DecodeError(str(node), value, "Value length is less than expected")

        pos, raw_head = 0, []
        for ctyp in node.ctypes:
            raw_head.append(value[pos : pos + ctyp.width])
            pos += ctyp.width

        if not node.is_dynamic:
            # no tail section
            return tuple(
                (
                    cls.decode(ctyp, val, **kwargs)
                    for ctyp, val in zip(node.ctypes, raw_head)
                )
            )

        ctyps_and_vals = list(zip(node.ctypes, raw_head))

        # ptrs are in the head section, convert them to ints in a single list
        ptrs = [
            int.from_bytes(val, "big")
            for ctyp, val in ctyps_and_vals
            if ctyp.is_dynamic
        ]
        # for each pointer copy the data from the dynamic section similar to array decoding
        data = deque([value[a:b] for a, b in zip(ptrs, ptrs[1:])] + [value[ptrs[-1] :]])
        # replace each ptr with its data - generator
        head = [
            data.popleft() if ctyp.is_dynamic else val for ctyp, val in ctyps_and_vals
        ]

        # return the decoded elements
        return tuple(
            [cls.decode(typ, val, **kwargs) for typ, val in zip(node.ctypes, head)]
        )


def _get_parser(schema: str):
    try:
        return _parsers[schema]
    except KeyError:
        _parsers[schema] = (ret := Parser.parse(schema))
        return ret


def abi_encode(schema: str, data: Any) -> bytes:
    return _ABIEncoder.encode(_get_parser(schema), data)


def abi_decode(schema: str, data: bytes) -> Any:
    return _ABIDecoder.decode(_get_parser(schema), data)


def is_abi_encodable(abi_type: str, data: Any) -> bool:
    try:
        abi_encode(abi_type, data)
        return True
    except ABIError:
        return False



import json
import time
from dataclasses import dataclass
from datetime import datetime, timedelta
from http import HTTPStatus
from typing import Callable, Generic, Optional, TypeVar

import requests

from boa.util.abi import Address
from boa.util.open_ctx import Open

DEFAULT_BLOCKSCOUT_URI = "https://eth.blockscout.com"
T = TypeVar("T")
P = TypeVar("P")


class ContractVerifier(Generic[T]):
    """A base class for contract verifiers.

    This class should be extended by specific verifiers like Etherscan, Blockscout, etc.
    """

    # Methods
    def verify(
        self,
        address: Address,
        contract_name: str,
        solc_json: dict,
        constructor_calldata: bytes,
        license_type: str = "1",
        wait: bool = False,
    ) -> Optional["VerificationResult[T]"]:
        raise NotImplementedError

    def wait_for_verification(self, identifier: T) -> None:
        raise NotImplementedError

    def is_verified(self, identifier: T) -> bool:
        raise NotImplementedError

    @staticmethod
    def _wait_until(
        predicate: Callable[[], P],
        wait_for: timedelta,
        backoff: timedelta,
        backoff_factor: float,
    ) -> P:
        timeout = datetime.now() + wait_for
        wait_time = backoff
        while datetime.now() < timeout:
            if result := predicate():
                return result
            time.sleep(wait_time.total_seconds())
            wait_time *= backoff_factor

        raise TimeoutError("Timeout waiting for verification to complete")


@dataclass
class Blockscout(ContractVerifier[Address]):
    """
    Allows users to verify contracts on Blockscout.
    This is independent of Vyper contracts, and can be used to verify any smart contract.
    """

    uri: str = DEFAULT_BLOCKSCOUT_URI
    api_key: Optional[str] = None
    timeout: timedelta = timedelta(minutes=2)
    backoff: timedelta = timedelta(milliseconds=500)
    backoff_factor: float = 1.1
    retry_http_codes: tuple[int, ...] = (
        HTTPStatus.NOT_FOUND,
        HTTPStatus.INTERNAL_SERVER_ERROR,
        HTTPStatus.SERVICE_UNAVAILABLE,
        HTTPStatus.GATEWAY_TIMEOUT,
    )

    def verify(
        self,
        address: Address,
        contract_name: str,
        solc_json: dict,
        constructor_calldata: bytes,
        license_type: str = "1",
        wait: bool = False,
    ) -> Optional["VerificationResult[Address]"]:
        """
        Verify the Vyper contract on Blockscout.
        :param address: The address of the contract.
        :param contract_name: The name of the contract.
        :param solc_json: The solc_json output of the Vyper compiler.
        :param constructor_calldata: The calldata for the constructor.
        :param license_type: The license to use for the contract. Defaults to "none".
        :param wait: Whether to return a VerificationResult immediately
                     or wait for verification to complete. Defaults to False
        """
        if license_type is None:
            license_type = "none"

        api_key = self.api_key or ""

        url = f"{self.uri}/api/v2/smart-contracts/{address}/"
        url += f"verification/via/vyper-standard-input?apikey={api_key}"
        data = {
            "compiler_version": solc_json["compiler_version"],
            "license_type": license_type,
        }
        files = {
            "files[0]": (
                contract_name,
                json.dumps(solc_json).encode("utf-8"),
                "application/json",
            )
        }

        response = requests.post(url, data=data, files=files)
        response.raise_for_status()
        print(response.json().get("message"))  # usually verification started

        if not wait:
            return VerificationResult(address, self)

        self.wait_for_verification(address)
        return None

    def wait_for_verification(self, address: Address) -> None:
        """
        Waits for the contract to be verified on Blockscout.
        :param address: The address of the contract.
        """
        self._wait_until(
            lambda: self.is_verified(address),
            self.timeout,
            self.backoff,
            self.backoff_factor,
        )
        msg = "Contract verified!"
        msg += f" {self.uri}/address/{address}?tab=contract_code"
        print(msg)

    def is_verified(self, address: Address) -> bool:
        api_key = self.api_key or ""
        url = f"{self.uri}/api/v2/smart-contracts/{address}?apikey={api_key}"

        response = requests.get(url)
        if response.status_code in self.retry_http_codes:
            return False
        response.raise_for_status()
        return response.json().get("is_verified", False)


_verifier: ContractVerifier = Blockscout()


@dataclass
class VerificationResult(Generic[T]):
    identifier: T
    verifier: ContractVerifier

    def wait_for_verification(self):
        self.verifier.wait_for_verification(self.identifier)

    def is_verified(self):
        return self.verifier.is_verified(self.identifier)


def _set_verifier(verifier):
    global _verifier
    _verifier = verifier


def get_verifier():
    global _verifier
    return _verifier


# TODO: maybe allow like `set_verifier("blockscout", *args, **kwargs)`
def set_verifier(verifier: ContractVerifier):
    return Open(get_verifier, _set_verifier, verifier)


def get_verification_bundle(contract_like):
    if not hasattr(contract_like, "deployer"):
        return None
    if not hasattr(contract_like.deployer, "solc_json"):
        return None
    return contract_like.deployer.solc_json


# should we also add a `verify_deployment` function?
def verify(
    contract, verifier: ContractVerifier = None, wait=False, **kwargs
) -> VerificationResult | None:
    """
    Verifies the contract on a block explorer.
    :param contract: The contract to verify.
    :param verifier: The block explorer verifier to use.
                     Defaults to get_verifier().
    :param wait: Whether to wait for verification to complete.
    """
    if verifier is None:
        verifier = get_verifier()

    if (bundle := get_verification_bundle(contract)) is None:
        raise ValueError(f"Not a contract! {contract}")

    # only VyperContract has this
    # TODO: expand VyperBlueprint and ABIContract
    ctor_calldata = getattr(contract, "ctor_calldata", b"")

    return verifier.verify(
        address=contract.address,
        solc_json=bundle,
        contract_name=contract.contract_name,
        constructor_calldata=ctor_calldata,
        wait=wait,
        **kwargs,
    )
